"""Connection module"""

from __future__ import annotations

import asyncio
import logging
from typing import Any, Callable

import aiohttp

from aiochroma.const import (
    CREDENTIALS,
    DEFAULT_PORT,
    DEFAULT_SLEEP,
    HEADERS,
    URL,
    URL_MAIN,
)
from aiochroma.error import ChromaError, ChromaResultError

_LOGGER = logging.getLogger(__name__)


class Connection:
    """AIOChroma connection"""

    def __init__(
        self,
        host: str,
        port: int = DEFAULT_PORT,
        session: aiohttp.ClientSession | None = None,
    ):
        """Properties for connection"""

        self._host = host
        self._port = port
        self._session: aiohttp.ClientSession = (
            session if session else aiohttp.ClientSession()
        )

        self._identity: dict[str, str] = ""
        self._connected: bool = False
        self._sid: int | None = None

    ### ------------------------
    ### Service methods -->
    ### ------------------------

    def _mark_disconnected(self) -> None:
        """Mark connection as disconnected."""

        self._connected = False

    ### ------------------------
    ### <-- Service methods
    ### ------------------------

    ### SEND REQUESTS ->

    async def async_request(
        self,
        endpoint: str,
        method: Callable,
        payload: str = "",
        interval: float = DEFAULT_SLEEP,
    ) -> dict[str, Any]:
        """Send a request"""

        # Check that we are connected
        if not self._connected and endpoint != URL_MAIN:
            if not await self.async_connect():
                raise ChromaError("Cannot connect to Chroma SDK")

        url = URL.format(
            self._host,
            self._port,
            endpoint if not self._sid else f"sid={self._sid}/{endpoint}",
        )

        try:
            async with method(
                url=url, headers=HEADERS, data=payload, ssl=True
            ) as response:
                responce_status = response.status
                if responce_status == 404:
                    raise ChromaError("Chroma SDK is not available")

                json_body = await response.json()

                if "result" in json_body and json_body["result"] != 0:
                    raise ChromaResultError(json_body["result"])

                await self.async_sleep(interval)

                return json_body

        except aiohttp.ClientConnectorError as ex:
            self._mark_disconnected()
            raise ChromaError("Cannot connect to the Chroma SDK") from ex

        except Exception as ex:
            self._mark_disconnected()
            raise ChromaError("Error communicating with the Chroma SDK") from ex

    async def async_delete(
        self,
        endpoint: str = "",
        payload: str = "",
        interval: float = DEFAULT_SLEEP,
    ) -> dict[str, Any]:
        """Send DELETE request"""

        return await self.async_request(
            endpoint, self._session.delete, payload, interval
        )

    async def async_get(
        self,
        endpoint: str,
        payload: str = "",
        interval: float = DEFAULT_SLEEP,
    ) -> dict[str, Any]:
        """Send GET request"""

        return await self.async_request(endpoint, self._session.get, payload, interval)

    async def async_post(
        self,
        endpoint: str,
        payload: str = "",
        interval: float = DEFAULT_SLEEP,
    ) -> dict[str, Any]:
        """Send POST request"""

        return await self.async_request(endpoint, self._session.post, payload, interval)

    async def async_put(
        self,
        endpoint: str,
        payload: str = "",
        interval: float = DEFAULT_SLEEP,
    ):
        """Send PUT request"""

        return await self.async_request(endpoint, self._session.put, payload, interval)

    ### <- SEND REQUESTS

    async def async_sleep(self, interval: float = DEFAULT_SLEEP) -> None:
        """Timeout"""

        await asyncio.sleep(interval)

    async def async_identify(self) -> None:
        """Identify Chroma API"""

        if not self._identity and not self._connected:
            result = await self.async_get(URL_MAIN)
            if "version" in result:
                self._identity = result
                _LOGGER.debug(f"Identity collected: {self._identity}")
            else:
                print(result)
                _LOGGER.warning("Cannot collect identity")

    async def async_connect(self) -> bool:
        """Connect to Chroma"""

        result = await self.async_post(URL_MAIN, CREDENTIALS)
        if "sessionid" in result:
            self._sid = int(result["sessionid"])
            self._connected = True
            return True
        return False

    async def async_disconnect(self) -> None:
        """Disconnect from Chroma"""

        await self.async_delete()

    async def async_keep(self) -> int:
        """Keep connection active"""

        result = await self.async_put("heartbeat")
        return result["tick"]

    @property
    def connected(self) -> bool:
        """Connection status"""

        return self._connected

    @property
    def sid(self) -> int:
        """Session ID"""

        return self._sid

    @property
    def identity(self) -> str:
        """API version"""

        return self._identity



"""Library init"""

from .aiochroma import AIOChroma
from .connection import Connection
from .dataclass import Color, Key
from .error import (
    ChromaError,
    ChromaResultError,
    ChromaUnknownLayout,
    ChromaUnknownTarget,
    ChromaWrongParameter,
)



"""Error module"""

from __future__ import annotations

from typing import Any, Optional


class ChromaError(Exception):
    """Base class for errors in AIOChroma library"""

    def __init__(
        self, *args: Any, message: Optional[str] = None, **_kwargs: Any
    ) -> None:
        """Initialise base error"""

        super().__init__(*args, message)


class ChromaUnknownLayout(ChromaError):
    """Unknown layout"""


class ChromaResultError(ChromaError):
    """Result error on the request"""


class ChromaUnknownTarget(ChromaError):
    """Unknown target"""


class ChromaWrongParameter(ChromaError):
    """Wrong parameter"""



"""AIOChroma module"""

from __future__ import annotations

import json
import logging
from typing import Any

import aiohttp

from .connection import Connection
from .const import (
    CHROMA_EFFECT,
    CHROMA_KEYBOARD_HEIGHT,
    CHROMA_KEYBOARD_WIDTH,
    CHROMA_PARAM,
    CHROMA_TARGETS,
    DEFAULT_BRIGHTNESS,
    DEFAULT_COLOR,
    DEFAULT_SLEEP,
    DEFAULT_SPACING,
    EFFECT_CUSTOM,
    EFFECT_NONE,
    EFFECT_STATIC,
    KEY_CODES,
    KEY_HEADSET,
    KEY_KEYBOARD,
    KEY_KEYPAD,
    KEY_LINK,
    KEY_MOUSE,
    KEY_MOUSEPAD,
    KEYREG,
    LAYOUT,
)
from .dataclass import Color
from .error import ChromaUnknownLayout, ChromaUnknownTarget, ChromaWrongParameter

_LOGGER = logging.getLogger(__name__)


class AIOChroma:
    """AIOChroma class"""

    def __init__(
        self,
        host: str,
        targets: list["str"],
        layout: str,
        session: aiohttp.ClientSession | None = None,
    ):
        """Initialize AIOChroma module"""

        if layout in LAYOUT:
            self._layout = LAYOUT[layout].copy()
        else:
            raise ChromaUnknownLayout(layout)

        self._connection: Connection = Connection(host, session=session)
        self._connected: bool = False

        self._state: dict[str, dict[str, Any]] = dict()

        for target in targets:
            if target in CHROMA_TARGETS:
                self._state[target] = dict()
                self._state[target]["brightness"] = DEFAULT_BRIGHTNESS
                self._state[target]["color"] = DEFAULT_COLOR
                self._state[target]["state"] = True

        self._state_keyboard: list[list[int]] = [
            [0 for i in range(CHROMA_KEYBOARD_WIDTH)]
            for j in range(CHROMA_KEYBOARD_HEIGHT)
        ]

    async def async_initialize(
        self, targets: list[str], color: Color = DEFAULT_COLOR
    ) -> bool:
        """Initialize devices"""

        for target in targets:
            if not target in CHROMA_TARGETS:
                raise ChromaUnknownTarget(target)

            await self.async_effect_color(target=target, color=color)
            _LOGGER.debug(f"Target `{target}` initialized")

    async def async_connect(self) -> bool:
        """Connect to Chroma"""

        # Identify first
        await self._connection.async_identify()

        # Connect
        self._connected = await self._connection.async_connect()

        # Because this is the only way to connect completely
        if self._connected:
            await self.async_keep()
            await self.async_keep()

        return self._connected

    async def async_keep(self) -> None:
        """Keep connection"""

        await self._connection.async_keep()

    async def async_disconnect(self) -> None:
        """Disconnect"""

        await self._connection.async_disconnect()
        self._connected = False

    ### EFFECTS -->

    async def async_effect_none(
        self, target: str, sleep: float = DEFAULT_SLEEP
    ) -> None:
        """Send none message"""

        if target not in CHROMA_TARGETS:
            raise ChromaUnknownTarget(target)

        payload = json.dumps(
            {
                CHROMA_EFFECT: EFFECT_NONE,
            }
        )

        await self._connection.async_put(
            endpoint=target, payload=payload, interval=sleep
        )
        await self.async_save_state(target=target, state=False)

    async def async_effect_color(
        self,
        target: str,
        color: Color | None = None,
        brightness: int | None = None,
        sleep: float = DEFAULT_SLEEP,
    ) -> None:
        """Single color effect"""

        if target not in CHROMA_TARGETS:
            raise ChromaUnknownTarget(target)

        if not brightness:
            brightness = self._state[target]["brightness"]
        if not color:
            color = self._state[target]["color"]

        payload = json.dumps(
            {
                CHROMA_EFFECT: EFFECT_STATIC,
                CHROMA_PARAM: color.scale(brightness).as_int(),
            }
        )

        await self._connection.async_put(
            endpoint=target, payload=payload, interval=sleep
        )
        await self.async_save_state(
            target=target, color=color, brightness=brightness, state=True
        )

    async def async_effect_blink(
        self,
        target: str,
        color: Color | None = None,
        brightness: int | None = None,
        repeats: int = 1,
        spacing: float = DEFAULT_SPACING,
    ) -> None:
        """Blink effect"""

        if target not in CHROMA_TARGETS:
            raise ChromaUnknownTarget(target)

        while repeats:
            await self.async_effect_none(target=target, sleep=spacing)
            await self.async_effect_color(
                target=target, color=color, brightness=brightness, sleep=spacing
            )

            repeats -= 1

        await self.async_effect_none(target=target, sleep=spacing)

    async def async_effect_keyboard(
        self, effect: list[list[int]], spacing: float = DEFAULT_SPACING
    ) -> None:
        """Keyboard by key effect"""

        payload = json.dumps(
            {
                CHROMA_EFFECT: EFFECT_CUSTOM,
                CHROMA_PARAM: effect,
            }
        )
        await self._connection.async_put(
            endpoint=KEY_KEYBOARD, payload=payload, interval=spacing
        )

        await self.async_save_state_keyboard(effect)

    ### <-- EFFECTS

    ### KEYBOARD -->

    async def async_parse_message(self, message) -> list[str]:
        """Parse message"""

        key_sp = dict()
        keys = list()

        # Find special keys
        for match in KEYREG.finditer(message):
            key_sp[match.start()] = match.group()

        # List the sequence of keys to use
        i = 0
        while i < len(message):
            if i in key_sp:
                keys.append(key_sp[i].replace("{", "").replace("}", ""))
                i += len(key_sp[i])
            else:
                keys.append(message[i])
                i += 1

        return keys

    async def async_set_keyboard_key(
        self,
        effect: list[list[int]],
        keys: list[str] | str,
        color: Color,
        brightness: int | None = None,
    ) -> list[list[int]]:
        """Set keyboard key value"""

        if type(keys) == str:
            keys = [keys]

        for item in keys:
            code = KEY_CODES[item]
            effect[code.row][code.column] = color.scale(brightness).as_int()

        return effect

    async def async_keyboard_sequence(
        self,
        message: str,
        color: Color = DEFAULT_COLOR,
        background: int = 0,
        brightness: int | None = None,
        tail: int = 0,
        repeats: int = 1,
        spacing: float = DEFAULT_SPACING,
        sleep: float = DEFAULT_SLEEP,
    ) -> None:
        """Send a key sequence"""

        _LOGGER.debug(
            f"Starting `keyboard_seqyence` with message=`{message}`, color=`{color}`, background=`{background}`, brightness=`{brightness}`, tail=`{tail}`, repeats=`{repeats}`, spacing=`{spacing}`, sleep=`{sleep}`"
        )

        # Get previous keyboard state
        was = self._state["keyboard"].copy()
        _LOGGER.debug(f"Previous keyboard state: {was}")

        # Parse the message
        keys = await self.async_parse_message(message)

        # Set background
        await self.async_effect_color(
            target=KEY_KEYBOARD, color=background, brightness=brightness
        )

        length = len(keys)
        # Don't allow tail longer than message
        if tail >= length:
            tail = length - 1
        while repeats:
            el = 0

            while el < length:
                effect = self._state_keyboard
                if keys[el] in self._layout:
                    # Turn off tail end
                    rudiment = el - tail - 1
                    if rudiment < length:
                        effect = await self.async_set_keyboard_key(
                            effect=effect,
                            keys=self._layout[keys[rudiment]],
                            color=background,
                            brightness=brightness,
                        )
                    effect = await self.async_set_keyboard_key(
                        effect=effect,
                        keys=self._layout[keys[el]],
                        color=color,
                        brightness=brightness,
                    )
                else:
                    _LOGGER.warning(f"Key '{keys[el]}' is unknown")
                    continue

                await self.async_effect_keyboard(effect=effect, spacing=spacing)

                el += 1

            repeats -= 1

            # Get rid of the tail in the end
            if repeats < 1:
                while el < length + tail + 1:
                    # Turn off tail end
                    rudiment = el - tail - 1
                    if rudiment > -1:
                        effect = await self.async_set_keyboard_key(
                            effect=effect,
                            keys=self._layout[keys[rudiment]],
                            color=background,
                            brightness=brightness,
                        )

                    await self.async_effect_keyboard(effect=effect, spacing=spacing)

                    el += 1

        # Recover previous state
        _LOGGER.debug(f"Recovering keyboard state: {was}")
        if was["state"]:
            await self.async_effect_color(
                target="keyboard", color=was["color"], brightness=was["brightness"]
            )
        else:
            await self.async_effect_none(target="keyboard")

    ### <-- KEYBOARD

    ### ACTIONS -->

    async def async_turn_off(self, target: str = "") -> None:
        """Turn off the target."""

        if not target in CHROMA_TARGETS:
            raise ChromaUnknownTarget(target)

        await self.async_effect_none(target=target)

        return

    async def async_turn_on(
        self,
        target: str = "",
        color: Color | None = None,
        brightness: int | None = None,
    ) -> None:
        """Turn on the target."""

        _LOGGER.debug(
            f"Received command turn_on with target=`{target}`, color=`{color}`, brightness=`{brightness}`"
        )

        if not target in CHROMA_TARGETS:
            raise ChromaUnknownTarget(target)

        if brightness is not None:
            if brightness > 255 or brightness < 0:
                raise ChromaWrongParameter(f"Wrong `brightness` value `{brightness}`")
            color = self._state[target]["color"]
            # _LOGGER.debug(
            #     f"Original color: r=`{color.r}`, g=`{color.g}`, b=`{color.b}`, brightness=`{color.brigtness()}`"
            # )

            # # Scale to new brightn
            # ratio = brightness / color.brigtness()
            # new_color = color.scale(ratio)

            # _LOGGER.debug(
            #     f"New color: r=`{new_color.r}`, g=`{new_color.g}`, b=`{new_color.b}`, brightness=`{new_color.brigtness()}`"
            # )

        elif color is None:
            color = self._state[target]["color"]

        await self.async_effect_color(target=target, color=color, brightness=brightness)

        return

    ### <-- ACTIONS

    ### STATES -->

    async def async_get_state(self, target: str = "") -> bool:
        """Get state of the target"""

        # Keep alive if asking
        await self.async_keep()

        if target == str():
            return self._state

        if not target in CHROMA_TARGETS:
            raise ChromaUnknownTarget(target)

        return self._state[target]

    async def async_save_state(
        self,
        target: str,
        color: Color | None = None,
        brightness: int | None = None,
        state: bool | None = None,
    ) -> None:
        """Save state of the target"""

        if not target in CHROMA_TARGETS:
            raise ChromaUnknownTarget(target)

        if color:
            if type(color) != Color:
                raise ValueError(f"Wrong color `{color}` of type `{type(color)}`")
            self._state[target]["color"] = color
            self._state[target]["state"] = True

        if brightness:
            if type(brightness) != int:
                raise ValueError(
                    f"Wrong state `{brightness}` of type `{type(brightness)}`"
                )
            self._state[target]["brightness"] = brightness
            self._state[target]["state"] = True

        if state is not None:
            if type(state) != bool:
                raise ValueError(f"Wrong state `{state}` of type `{type(state)}`")
            self._state[target]["state"] = state

    async def async_save_state_keyboard(self, state: list[list[int]] | int = 0) -> None:
        """Save state of target"""

        if (
            type(state) == list
            and len(state) == CHROMA_KEYBOARD_HEIGHT
            and type(state[0]) == list
            and len(state[0]) == CHROMA_KEYBOARD_WIDTH
        ):
            self._state_keyboard = state
            return

        # Save the color
        if type(state) == int:
            self._state_keyboard = [
                [state for i in range(CHROMA_KEYBOARD_WIDTH)]
                for j in range(CHROMA_KEYBOARD_HEIGHT)
            ]
        else:
            raise ValueError(f"Wrong value `{state}` of type `{type(state)}`")

    ### <-- STATES

    @property
    def identity(self) -> dict[str, str]:
        """Chroma versions"""

        return self._connection.identity

    @property
    def connected(self) -> bool:
        """Connection state"""

        return self._connection.connected



"""Dataclass module for AIOChroma"""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Key:
    """Key class"""

    row: int
    column: int


@dataclass
class Color:
    """Color class"""

    r: int = 0
    g: int = 0
    b: int = 0

    def as_int(self):
        """Return single-int number"""

        return self.r + self.g * 256 + self.b * 65536

    def brigtness(self) -> int:
        """Return maximum brightness in single channel"""

        return max(self.r, self.g, self.b)

    def scale(self, scale: float | int) -> Color:
        """Return scaled color"""

        if type(scale) == float:
            return Color(
                round(self.r * scale), round(self.g * scale), round(self.b * scale)
            )
        elif type(scale) == int:
            return self.scale(scale / 255)
        else:
            raise ValueError(f"Unknown scale: `{scale}`")
